package cn.ecpark.service.usergw.config;

import java.io.InputStream;
import java.lang.reflect.Method;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;

import org.apache.dubbo.config.ReferenceConfig;
import org.apache.dubbo.rpc.service.GenericService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.SpringApplication;
import org.springframework.cloud.gateway.filter.FilterDefinition;
import org.springframework.cloud.gateway.handler.predicate.PredicateDefinition;
import org.springframework.cloud.gateway.route.RouteDefinition;
import org.springframework.cloud.gateway.route.RouteDefinitionLocator;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.stereotype.Component;
import org.springframework.stereotype.Controller;
import org.yaml.snakeyaml.Yaml;

import cn.ecpark.service.usergw.biz.filters.bean.GenericServicePool;
import cn.ecpark.service.usergw.biz.filters.factory.DubboGatewayFilterFactory;
import cn.ecpark.service.usergw.impl.http.Http2Dubbo;
import cn.ecpark.service.usergw.utils.Convert;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Flux;

/**
 * ConfigGateway
 */
@Component
@Controller
@Slf4j
public class ConfigGateway implements RouteDefinitionLocator {

  // List<MediaType> mediaTypes = new ArrayList<>();

  List<FilterDefinition> defaultFilters = new ArrayList<>();
  HashMap<String, BiConsumer<ReferenceConfig<GenericService> , String>> specialField = new HashMap<String,  BiConsumer<ReferenceConfig<GenericService> , String>>();

  @Autowired
  private ApplicationContext appContext;

  @Bean
  public GenericServicePool genericServicePool() {
    return new GenericServicePool();
  }

  @Bean
  public DubboGatewayFilterFactory dubboFilterFactory() {
    return new DubboGatewayFilterFactory();
  }


  public ConfigGateway() {
    specialField.put("application", ConfigSpecialFunction::setApplication);
  }

  @Override
  @SuppressWarnings("unchecked")
  public Flux<RouteDefinition> getRouteDefinitions() {
    // WebFluxConfigurationSupport a;

    Object inputStream = null;
    String[] gatewayConfigPathList = { "gateway.yml", "gateway.yaml", "Gateway.yml", "Gateway.yaml" };
    for (String gatewayConfigPath : gatewayConfigPathList) {
      inputStream = this.getClass().getClassLoader().getResourceAsStream(gatewayConfigPath);
      if (inputStream != null) {
        break;
      }
    }

    if (inputStream != null) {
      Map<String, Object> configYaml = new Yaml().load((InputStream) inputStream);
      if (configYaml != null) {

        // streaming-media-types 这个先不管
        // NettyWriteResponseFilter nwrf =
        // applicationContext.getBean(NettyWriteResponseFilter.class);
        // try {
        // Field field = nwrf.getClass().getDeclaredField("streamingMediaTypes");
        // field.setAccessible(true);
        // List<MediaType> mts = new ArrayList<MediaType>();
        // mts.add(MediaType.TEXT_EVENT_STREAM);
        // mts.add(MediaType.APPLICATION_STREAM_JSON);
        // field.set(nwrf, mts);
        // } catch (Exception e) {
        // log.error(e.toString());
        // }

        Map<String, Object> defaultYaml = (Map<String, Object>) configYaml.get("default");
        Map<String, Object> dubboYaml = (Map<String, Object>) configYaml.get("dubbo");

        List<RouteDefinition> routeList = new ArrayList<RouteDefinition>();

        if (defaultYaml != null && !defaultYaml.isEmpty()) {
          this.getDefaultFilter(defaultFilters, defaultYaml);
          this.configDefault(routeList, defaultYaml);
        }

        if (dubboYaml != null && !dubboYaml.isEmpty()) {
          this.configHttp2Dubbo(routeList, dubboYaml);
        }

        if (!routeList.isEmpty()) {
          return Flux.fromIterable(routeList);
        }

      }
    }

    return Flux.empty();
    // return Flux.fromIterable(it);
  }

  @SuppressWarnings("unchecked")
  private void getDefaultFilter(List<FilterDefinition> filters, Map<String, Object> defaultYaml) {
    // default-filters: 下的相关设置
    Object unknownDefaultFilters = defaultYaml.get("default-filters");
    if (unknownDefaultFilters != null) {
      List<String> defaultFiltersYaml = (ArrayList<String>) unknownDefaultFilters;
      for (String filterString : defaultFiltersYaml) {
        filters.add(new FilterDefinition(filterString));
      }
    }
  }

  @SuppressWarnings("unchecked")
  private void configDefault(List<RouteDefinition> routeList, Map<String, Object> defaultYaml) {
    if (defaultYaml != null) {

      Object unknownRoutes = defaultYaml.get("routes");
      if (unknownRoutes != null) {
        List<LinkedHashMap<String, List<String>>> routes = (ArrayList<LinkedHashMap<String, List<String>>>) unknownRoutes;
        for (LinkedHashMap<String, List<String>> iter : routes) {

          List<FilterDefinition> filters = new ArrayList<>();
          List<PredicateDefinition> predicates = new ArrayList<>();
          RouteDefinition rd = new RouteDefinition();

          // 设置基础属性
          this.ParseAndSetBase(rd, iter);

          // predicates: 下的相关属性
          this.ParseAndAddPredicates(predicates, iter, "predicates");

          // filters: 下的相关属性
          this.ParseAndAddFilters(filters, iter, "filters");

          rd.setPredicates(predicates);
          rd.setFilters(filters);

          routeList.add(rd);
        }
      }
    }
  }

  @SuppressWarnings("unchecked")
  private void configHttp2Dubbo(List<RouteDefinition> routeList, Map<String, Object> configDubbo) {

    try {

      Object unknownRoutes = configDubbo.get("routes");
      if (unknownRoutes != null) {

        List<LinkedHashMap<String, List<String>>> routes = (ArrayList<LinkedHashMap<String, List<String>>>) unknownRoutes;
        for (LinkedHashMap<String, List<String>> iter : routes) {

          List<FilterDefinition> filters = new ArrayList<>();
          List<PredicateDefinition> predicates = new ArrayList<>();
          RouteDefinition rd = new RouteDefinition();

          // 设置基础属性
          String uri = this.ParseDubboUriAndSetBase(rd, iter);

          // predicates: 下的相关属性
          this.ParseAndAddPredicates(predicates, iter, "predicates");

          // filters: 下的相关属性
          this.ParseAndAddDubboFilters(filters, iter, "filters");

          rd.setPredicates(predicates);
          rd.setFilters(filters);

          routeList.add(rd);

        }

      }

    } catch (Exception e) {
      // TODO: 任何错误都直接终止退出
      log.error(e.toString());
      SpringApplication.exit(appContext);
    }

    // gs.$invoke(method, parameterTypes, args)

  }

  private void ParseAndSetBase(RouteDefinition rd, LinkedHashMap<String, List<String>> iter) {
    // 设置id
    Object id = iter.get("id");
    if (id != null) {
      rd.setId((String) id);
    }

    // 设置uri
    Object uri = iter.get("uri");
    if (uri != null) {
      rd.setUri(URI.create((String) uri));
    }

    // 设置uri
    Object order = iter.get("order");
    if (order != null) {
      rd.setOrder((int) order);
    }
  }

  private String ParseDubboUriAndSetBase(RouteDefinition rd, LinkedHashMap<String, List<String>> iter) {
    // 设置id
    Object id = iter.get("id");
    if (id != null) {
      rd.setId((String) id);
    }

    // 设置uri
    Object uri = iter.get("uri");
    if(uri != null) {
      try {
        rd.setUri(URI.create((String) uri));
      } catch (Exception e) {
        log.warn("dubbo Uri error");
        rd.setUri(URI.create("dubbo://yame"));
      }
    } else {
      rd.setUri(URI.create("dubbo://yame"));
    }

    // 设置uri
    Object order = iter.get("order");
    if (order != null) {
      rd.setOrder((int) order);
    }

    if (uri != null) {
      String uriString = (String) uri;
      uriString = uriString.trim();
      if (uriString.startsWith("dubbo://")) { // dubbo://127.0.0.1/interface?
        return uriString;
      }
    }



    // GenericServicePool gsPool = appContext.getBean(GenericServicePool.class);
    // GenericService gs = gsPool.get(rkey);
    // if(gs == null) {
    //     ReferenceConfig<GenericService> reference = new ReferenceConfig<GenericService>(); // 该实例很重量，里面封装了所有与注册中心及服务提供方连接，请缓存
    //     reference.setApplication(new ApplicationConfig("dubbo-exchange")); // 必须
    //     // reference.setRegistry(new RegistryConfig("zookeeper://127.0.0.1:2181"));
    //     reference.setGroup("test");
    //     reference.setInterface("ocean.demo.api.IExchange"); // 弱类型接口名 必须
    //     reference.setVersion("1.0.0"); // 必须
    //     reference.setConnections(rconnections);
    //     reference.setGeneric(true); // 声明为泛化接口
    //     // gsPool.add(rkey, reference.get());
    //     gs = reference.get();
    //     gsPool.put(rkey, gs);
    // }

    ReferenceConfig<GenericService> reference = new ReferenceConfig<GenericService>();
    String UriString = "dubbo://"; 

    Object application = iter.get("application");
    if (application != null) {
      BiConsumer<ReferenceConfig<GenericService>, String> doFunc = specialField.get("application");
      if(doFunc != null) {
        doFunc.accept(reference, (String)application); 
      }
      UriString += (String)application + "/";
    }

    Object group = iter.get("group");
    if (group != null) {
      BiConsumer<ReferenceConfig<GenericService>, String> doFunc = specialField.get((String)group);
      if(doFunc != null) {
        doFunc.accept(reference, (String)application); 
      } else {
        try {
          ConfigSpecialFunction.setDefault(reference, "group", (String)group);
        } catch (Exception e) {
          log.error("配置解析错误 字段: {}:{}", "group", group);
        }
        
      }
      UriString += (String)application + "/";      
    }

    Object registry = iter.get("registry");
    if (registry != null) {
      rd.setOrder((int) registry);
    }





    return "";
  }

  private void ParseAndAddPredicates(List<PredicateDefinition> predicates, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> predicatesYaml = iter.get(yamlField);
    if (predicatesYaml != null) {
      for (String predicateString : predicatesYaml) {
        predicates.add(new PredicateDefinition(predicateString));
      }
    }
  }

  private void ParseAndAddFilters(List<FilterDefinition> filters, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> filtersYaml = iter.get(yamlField);
    if (filtersYaml != null) {
      filters.addAll(defaultFilters);
      for (String filterString : filtersYaml) {
        filters.add(new FilterDefinition(filterString));
      }
    }
  }

  private void ParseAndAddDubboFilters(List<FilterDefinition> filters, LinkedHashMap<String, List<String>> iter,
      String yamlField) {
    List<String> filtersYaml = iter.get(yamlField);
    if (filtersYaml != null) {
      filters.addAll(defaultFilters);
      for (String filterString : filtersYaml) {
        FilterDefinition fd = new FilterDefinition(filterString);
        log.info(fd.getName());
        if (!fd.getName().equals("Dubbo")) {
          filters.add(fd);
        } else {
          filters.add(fd);
        }
      }
    }
  }

  private static void callMethod(ReferenceConfig ref, String name, String params) throws Exception {
    Method method = ref.getClass().getMethod("set" + Convert.firstUpperCase(name));
    method.invoke(ref, method.getParameterTypes(), params);
  }
}